@@ -6,6 +6,7 @@ import (
 	"io"
 	"net"
 	"strings"
+	"sync"
 	"time"
 
 	iradix "github.com/hashicorp/go-immutable-radix"
@@ -18,6 +19,13 @@ type PluginBlockIP struct {
 	blockedIPs      map[string]interface{}
 	logger          io.Writer
 	format          string
+
+	// Hot-reloading support
+	rwLock          sync.RWMutex
+	configFile      string
+	configWatcher   *ConfigWatcher
+	stagingPrefixes *iradix.Tree
+	stagingIPs      map[string]interface{}
 }
 
 func (plugin *PluginBlockIP) Name() string {
@@ -29,24 +37,44 @@ func (plugin *PluginBlockIP) Description() string {
 }
 
 func (plugin *PluginBlockIP) Init(proxy *Proxy) error {
-	dlog.Noticef("Loading the set of IP blocking rules from [%s]", proxy.blockIPFile)
-	bin, err := ReadTextFile(proxy.blockIPFile)
+	plugin.configFile = proxy.blockIPFile
+	dlog.Noticef("Loading the set of IP blocking rules from [%s]", plugin.configFile)
+
+	lines, err := ReadTextFile(plugin.configFile)
 	if err != nil {
 		return err
 	}
+
 	plugin.blockedPrefixes = iradix.New()
 	plugin.blockedIPs = make(map[string]interface{})
-	for lineNo, line := range strings.Split(string(bin), "\n") {
+
+	if err := plugin.loadRules(lines, plugin.blockedPrefixes, plugin.blockedIPs); err != nil {
+		return err
+	}
+
+	if len(proxy.blockIPLogFile) > 0 {
+		plugin.logger = Logger(proxy.logMaxSize, proxy.logMaxAge, proxy.logMaxBackups, proxy.blockIPLogFile)
+		plugin.format = proxy.blockIPFormat
+	}
+
+	return nil
+}
+
+// loadRules parses and loads IP rules into the provided tree and map
+func (plugin *PluginBlockIP) loadRules(lines string, prefixes *iradix.Tree, ips map[string]interface{}) error {
+	for lineNo, line := range strings.Split(lines, "\n") {
 		line = TrimAndStripInlineComments(line)
 		if len(line) == 0 {
 			continue
 		}
+
 		ip := net.ParseIP(line)
 		trailingStar := strings.HasSuffix(line, "*")
 		if len(line) < 2 || (ip != nil && trailingStar) {
 			dlog.Errorf("Suspicious IP blocking rule [%s] at line %d", line, lineNo)
 			continue
 		}
+
 		if trailingStar {
 			line = line[:len(line)-1]
 		}
@@ -61,39 +89,111 @@ func (plugin *PluginBlockIP) Init(proxy *Proxy) error {
 			dlog.Errorf("Invalid rule: [%s] - wildcards can only be used as a suffix at line %d", line, lineNo)
 			continue
 		}
+
 		line = strings.ToLower(line)
 		if trailingStar {
-			plugin.blockedPrefixes, _, _ = plugin.blockedPrefixes.Insert([]byte(line), 0)
+			var updated *iradix.Tree
+			updated, _, _ = prefixes.Insert([]byte(line), 0)
+			prefixes = updated
 		} else {
-			plugin.blockedIPs[line] = true
+			ips[line] = true
 		}
 	}
-	if len(proxy.blockIPLogFile) == 0 {
-		return nil
-	}
-	plugin.logger = Logger(proxy.logMaxSize, proxy.logMaxAge, proxy.logMaxBackups, proxy.blockIPLogFile)
-	plugin.format = proxy.blockIPFormat
 
 	return nil
 }
 
 func (plugin *PluginBlockIP) Drop() error {
+	if plugin.configWatcher != nil {
+		plugin.configWatcher.RemoveFile(plugin.configFile)
+	}
 	return nil
 }
 
-func (plugin *PluginBlockIP) Reload() error {
+// PrepareReload loads new rules into staging structures but doesn't apply them yet
+func (plugin *PluginBlockIP) PrepareReload() error {
+	// Read the configuration file
+	lines, err := SafeReadTextFile(plugin.configFile)
+	if err != nil {
+		return fmt.Errorf("error reading config file during reload preparation: %w", err)
+	}
+
+	// Create staging structures
+	plugin.stagingPrefixes = iradix.New()
+	plugin.stagingIPs = make(map[string]interface{})
+
+	// Load rules into staging structures
+	if err := plugin.loadRules(lines, plugin.stagingPrefixes, plugin.stagingIPs); err != nil {
+		return fmt.Errorf("error parsing config during reload preparation: %w", err)
+	}
+
 	return nil
 }
 
+// ApplyReload atomically replaces the active rules with the staging ones
+func (plugin *PluginBlockIP) ApplyReload() error {
+	if plugin.stagingPrefixes == nil || plugin.stagingIPs == nil {
+		return errors.New("no staged configuration to apply")
+	}
+
+	// Use write lock to swap rule structures
+	plugin.rwLock.Lock()
+	plugin.blockedPrefixes = plugin.stagingPrefixes
+	plugin.blockedIPs = plugin.stagingIPs
+	plugin.stagingPrefixes = nil
+	plugin.stagingIPs = nil
+	plugin.rwLock.Unlock()
+
+	dlog.Noticef("Applied new configuration for plugin [%s]", plugin.Name())
+	return nil
+}
+
+// CancelReload cleans up any staging resources
+func (plugin *PluginBlockIP) CancelReload() {
+	plugin.stagingPrefixes = nil
+	plugin.stagingIPs = nil
+}
+
+// Reload implements hot-reloading for the plugin
+func (plugin *PluginBlockIP) Reload() error {
+	dlog.Noticef("Reloading configuration for plugin [%s]", plugin.Name())
+
+	// Prepare the new configuration
+	if err := plugin.PrepareReload(); err != nil {
+		plugin.CancelReload()
+		return err
+	}
+
+	// Apply the new configuration
+	return plugin.ApplyReload()
+}
+
+// GetConfigPath returns the path to the plugin's configuration file
+func (plugin *PluginBlockIP) GetConfigPath() string {
+	return plugin.configFile
+}
+
+// SetConfigWatcher sets the config watcher for this plugin
+func (plugin *PluginBlockIP) SetConfigWatcher(watcher *ConfigWatcher) {
+	plugin.configWatcher = watcher
+}
+
 func (plugin *PluginBlockIP) Eval(pluginsState *PluginsState, msg *dns.Msg) error {
 	if pluginsState.sessionData["whitelisted"] != nil {
 		return nil
 	}
+
 	answers := msg.Answer
 	if len(answers) == 0 {
 		return nil
 	}
+
 	reject, reason, ipStr := false, "", ""
+
+	// Use read lock for thread-safe access to configuration
+	plugin.rwLock.RLock()
+	defer plugin.rwLock.RUnlock()
+
 	for _, answer := range answers {
 		header := answer.Header()
 		Rrtype := header.Rrtype
@@ -117,16 +217,21 @@ func (plugin *PluginBlockIP) Eval(pluginsState *PluginsState, msg *dns.Msg) erro
 			}
 		}
 	}
+
 	if reject {
 		pluginsState.action = PluginsActionReject
 		pluginsState.returnCode = PluginsReturnCodeReject
 		if plugin.logger != nil {
 			qName := pluginsState.qName
 			var clientIPStr string
-			if pluginsState.clientProto == "udp" {
+			switch pluginsState.clientProto {
+			case "udp":
 				clientIPStr = (*pluginsState.clientAddr).(*net.UDPAddr).IP.String()
-			} else {
+			case "tcp", "local_doh":
 				clientIPStr = (*pluginsState.clientAddr).(*net.TCPAddr).IP.String()
+			default:
+				// Ignore internal flow.
+				return nil
 			}
 			var line string
 			if plugin.format == "tsv" {
