@@ -6,6 +6,7 @@ import (
 	"io"
 	"net"
 	"strings"
+	"sync"
 	"time"
 
 	"github.com/jedisct1/dlog"
@@ -21,7 +22,11 @@ type BlockedNames struct {
 
 const aliasesLimit = 8
 
-var blockedNames *BlockedNames
+var (
+	// protects access to the blockedNames global variable
+	blockedNamesLock sync.RWMutex
+	blockedNames     *BlockedNames
+)
 
 func (blockedNames *BlockedNames) check(pluginsState *PluginsState, qName string, aliasFor *string) (bool, error) {
 	reject, reason, xweeklyRanges := blockedNames.patternMatcher.Eval(qName)
@@ -44,10 +49,14 @@ func (blockedNames *BlockedNames) check(pluginsState *PluginsState, qName string
 	pluginsState.returnCode = PluginsReturnCodeReject
 	if blockedNames.logger != nil {
 		var clientIPStr string
-		if pluginsState.clientProto == "udp" {
+		switch pluginsState.clientProto {
+		case "udp":
 			clientIPStr = (*pluginsState.clientAddr).(*net.UDPAddr).IP.String()
-		} else {
+		case "tcp", "local_doh":
 			clientIPStr = (*pluginsState.clientAddr).(*net.TCPAddr).IP.String()
+		default:
+			// Ignore internal flow.
+			return false, nil
 		}
 		var line string
 		if blockedNames.format == "tsv" {
@@ -72,6 +81,10 @@ func (blockedNames *BlockedNames) check(pluginsState *PluginsState, qName string
 // ---
 
 type PluginBlockName struct {
+	// Hot-reloading support
+	configFile     string
+	configWatcher  *ConfigWatcher
+	stagingBlocked *BlockedNames
 }
 
 func (plugin *PluginBlockName) Name() string {
@@ -83,76 +96,181 @@ func (plugin *PluginBlockName) Description() string {
 }
 
 func (plugin *PluginBlockName) Init(proxy *Proxy) error {
-	dlog.Noticef("Loading the set of blocking rules from [%s]", proxy.blockNameFile)
-	bin, err := ReadTextFile(proxy.blockNameFile)
+	plugin.configFile = proxy.blockNameFile
+	dlog.Noticef("Loading the set of blocking rules from [%s]", plugin.configFile)
+
+	lines, err := ReadTextFile(plugin.configFile)
 	if err != nil {
 		return err
 	}
+
 	xBlockedNames := BlockedNames{
 		allWeeklyRanges: proxy.allWeeklyRanges,
 		patternMatcher:  NewPatternMatcher(),
 	}
-	for lineNo, line := range strings.Split(string(bin), "\n") {
+
+	if err := plugin.loadRules(lines, &xBlockedNames); err != nil {
+		return err
+	}
+
+	if len(proxy.blockNameLogFile) > 0 {
+		xBlockedNames.logger = Logger(proxy.logMaxSize, proxy.logMaxAge, proxy.logMaxBackups, proxy.blockNameLogFile)
+		xBlockedNames.format = proxy.blockNameFormat
+	}
+
+	blockedNamesLock.Lock()
+	blockedNames = &xBlockedNames
+	blockedNamesLock.Unlock()
+
+	return nil
+}
+
+// loadRules parses and loads name patterns into the BlockedNames
+func (plugin *PluginBlockName) loadRules(lines string, blockedNamesObj *BlockedNames) error {
+	for lineNo, line := range strings.Split(lines, "\n") {
 		line = TrimAndStripInlineComments(line)
 		if len(line) == 0 {
 			continue
 		}
-		parts := strings.Fields(line)
+
+		// Handle time-based restrictions with @timerange format
+		parts := strings.Split(line, "@")
 		timeRangeName := ""
 		if len(parts) == 2 {
-			if timeRangeParts := strings.Split(parts[1], "@"); len(timeRangeParts) == 2 {
-				timeRangeName = strings.TrimSpace(timeRangeParts[1])
-			} else {
-				dlog.Errorf("Syntax error in block rules at line %d", 1+lineNo)
-				continue
-			}
+			line = strings.TrimSpace(parts[0])
+			timeRangeName = strings.TrimSpace(parts[1])
 		} else if len(parts) > 2 {
-			dlog.Errorf("Syntax error in block rules at line %d", 1+lineNo)
+			dlog.Errorf("Syntax error in block rules at line %d -- Unexpected @ character", 1+lineNo)
 			continue
 		}
+
+		// Look up the time range if specified
 		var weeklyRanges *WeeklyRanges
 		if len(timeRangeName) > 0 {
-			weeklyRangesX, ok := (*xBlockedNames.allWeeklyRanges)[timeRangeName]
+			weeklyRangesX, ok := (*blockedNamesObj.allWeeklyRanges)[timeRangeName]
 			if !ok {
 				dlog.Errorf("Time range [%s] not found at line %d", timeRangeName, 1+lineNo)
 			} else {
 				weeklyRanges = &weeklyRangesX
 			}
 		}
-		if err := xBlockedNames.patternMatcher.Add(line, weeklyRanges, lineNo+1); err != nil {
+		if err := blockedNamesObj.patternMatcher.Add(line, weeklyRanges, lineNo+1); err != nil {
 			dlog.Error(err)
 			continue
 		}
 	}
-	blockedNames = &xBlockedNames
-	if len(proxy.blockNameLogFile) == 0 {
-		return nil
-	}
-	blockedNames.logger = Logger(proxy.logMaxSize, proxy.logMaxAge, proxy.logMaxBackups, proxy.blockNameLogFile)
-	blockedNames.format = proxy.blockNameFormat
 
 	return nil
 }
 
 func (plugin *PluginBlockName) Drop() error {
+	if plugin.configWatcher != nil {
+		plugin.configWatcher.RemoveFile(plugin.configFile)
+	}
 	return nil
 }
 
-func (plugin *PluginBlockName) Reload() error {
+// PrepareReload loads new patterns into staging structure but doesn't apply them yet
+func (plugin *PluginBlockName) PrepareReload() error {
+	// Read the configuration file
+	lines, err := SafeReadTextFile(plugin.configFile)
+	if err != nil {
+		return fmt.Errorf("error reading config file during reload preparation: %w", err)
+	}
+
+	// Get current BlockedNames to access allWeeklyRanges and log settings
+	blockedNamesLock.RLock()
+	currentBlockedNames := blockedNames
+	blockedNamesLock.RUnlock()
+
+	if currentBlockedNames == nil {
+		return errors.New("no existing blocked names configuration to base reload on")
+	}
+
+	// Create staging structure
+	plugin.stagingBlocked = &BlockedNames{
+		allWeeklyRanges: currentBlockedNames.allWeeklyRanges,
+		patternMatcher:  NewPatternMatcher(),
+		logger:          currentBlockedNames.logger,
+		format:          currentBlockedNames.format,
+	}
+
+	// Load rules into staging structure
+	if err := plugin.loadRules(lines, plugin.stagingBlocked); err != nil {
+		return fmt.Errorf("error parsing config during reload preparation: %w", err)
+	}
+
 	return nil
 }
 
+// ApplyReload atomically replaces the active rules with the staging ones
+func (plugin *PluginBlockName) ApplyReload() error {
+	if plugin.stagingBlocked == nil {
+		return errors.New("no staged configuration to apply")
+	}
+
+	// Use write lock to swap rule structures
+	blockedNamesLock.Lock()
+	blockedNames = plugin.stagingBlocked
+	blockedNamesLock.Unlock()
+
+	plugin.stagingBlocked = nil
+
+	dlog.Noticef("Applied new configuration for plugin [%s]", plugin.Name())
+	return nil
+}
+
+// CancelReload cleans up any staging resources
+func (plugin *PluginBlockName) CancelReload() {
+	plugin.stagingBlocked = nil
+}
+
+// Reload implements hot-reloading for the plugin
+func (plugin *PluginBlockName) Reload() error {
+	dlog.Noticef("Reloading configuration for plugin [%s]", plugin.Name())
+
+	// Prepare the new configuration
+	if err := plugin.PrepareReload(); err != nil {
+		plugin.CancelReload()
+		return err
+	}
+
+	// Apply the new configuration
+	return plugin.ApplyReload()
+}
+
+// GetConfigPath returns the path to the plugin's configuration file
+func (plugin *PluginBlockName) GetConfigPath() string {
+	return plugin.configFile
+}
+
+// SetConfigWatcher sets the config watcher for this plugin
+func (plugin *PluginBlockName) SetConfigWatcher(watcher *ConfigWatcher) {
+	plugin.configWatcher = watcher
+}
+
 func (plugin *PluginBlockName) Eval(pluginsState *PluginsState, msg *dns.Msg) error {
-	if blockedNames == nil || pluginsState.sessionData["whitelisted"] != nil {
+	if pluginsState.sessionData["whitelisted"] != nil {
 		return nil
 	}
-	_, err := blockedNames.check(pluginsState, pluginsState.qName, nil)
+
+	blockedNamesLock.RLock()
+	localBlockedNames := blockedNames
+	blockedNamesLock.RUnlock()
+
+	if localBlockedNames == nil {
+		return nil
+	}
+
+	_, err := localBlockedNames.check(pluginsState, pluginsState.qName, nil)
 	return err
 }
 
 // ---
 
 type PluginBlockNameResponse struct {
+	// The response plugin doesn't need any special fields for hot-reloading
+	// as it uses the shared blockedNames
 }
 
 func (plugin *PluginBlockNameResponse) Name() string {
@@ -172,13 +290,24 @@ func (plugin *PluginBlockNameResponse) Drop() error {
 }
 
 func (plugin *PluginBlockNameResponse) Reload() error {
+	// The response plugin doesn't need to reload anything itself
+	// as it uses the shared blockedNames that is reloaded by PluginBlockName
 	return nil
 }
 
 func (plugin *PluginBlockNameResponse) Eval(pluginsState *PluginsState, msg *dns.Msg) error {
-	if blockedNames == nil || pluginsState.sessionData["whitelisted"] != nil {
+	if pluginsState.sessionData["whitelisted"] != nil {
 		return nil
 	}
+
+	blockedNamesLock.RLock()
+	localBlockedNames := blockedNames
+	blockedNamesLock.RUnlock()
+
+	if localBlockedNames == nil {
+		return nil
+	}
+
 	aliasFor := pluginsState.qName
 	aliasesLeft := aliasesLimit
 	answers := msg.Answer
@@ -201,7 +330,7 @@ func (plugin *PluginBlockNameResponse) Eval(pluginsState *PluginsState, msg *dns
 		if err != nil {
 			return err
 		}
-		if blocked, err := blockedNames.check(pluginsState, target, &aliasFor); blocked || err != nil {
+		if blocked, err := localBlockedNames.check(pluginsState, target, &aliasFor); blocked || err != nil {
 			return err
 		}
 		aliasesLeft--
